// ---------------------------------------------------------------------------
// - InputOutput.cpp                                                         -
// - standard object library - input-output stream class implementation      -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2005 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Vector.hpp"
#include "Boolean.hpp"
#include "Integer.hpp"
#include "Character.hpp"
#include "Exception.hpp"
#include "InputOutput.hpp"

namespace afnix {

  // the input supported quarks
  static const long QUARK_EOFP    = String::intern ("eof-p");
  static const long QUARK_READ    = String::intern ("read");
  static const long QUARK_PUSHB   = String::intern ("pushback");
  static const long QUARK_VALIDP  = String::intern ("valid-p");
  static const long QUARK_READLN  = String::intern ("readln");
  static const long QUARK_BUFLEN  = String::intern ("get-buffer-length");

  // the output supported quarks
  static const long QUARK_WRITE   = String::intern ("write");
  static const long QUARK_WRITELN = String::intern ("writeln");
  static const long QUARK_ERRORLN = String::intern ("errorln");
  static const long QUARK_NEWLINE = String::intern ("newline");

  // the input-output supported quarks
  static const long QUARK_RESET   = String::intern ("reset");

  // create a new input-output stream

  InputOutput::InputOutput (void) {
    d_buffer.setrflg (false);
    reset ();
  }

  // create a new input-output stream with a data string

  InputOutput::InputOutput (const String& data) {
    d_buffer.setrflg (false);
    reset ();
    d_buffer.add (data);
  }
  
  // create a new input-output stream with a buffer and a size

  InputOutput::InputOutput (const t_byte* data, const long size) {
    d_buffer.setrflg (false);
    reset ();
    d_buffer.add ((const char*) data, size);
  }

  // reset this input-output stream

  void InputOutput::reset (void) {
    wrlock ();
    d_buffer.reset ();
    unlock ();
  }

  // return the class name

  String InputOutput::repr (void) const {
    return "InputOutput";
  }

  // read a character from the stream

  char InputOutput::read (void) {
    rdlock ();
    char result = d_buffer.isempty () ? eofc : d_buffer.read ();
    unlock ();
    return result;
  }

  // return the eof flag

  bool InputOutput::iseof (void) const {
    rdlock ();
    bool result = d_buffer.isempty ();
    unlock ();
    return result;
  }

  // return true if we can read a valid character

  bool InputOutput::valid (const long tout) const {
    rdlock ();
    bool result = (d_buffer.isempty () == false);
    unlock ();
    return result;
  }

  // write a character in the buffer

  void InputOutput::write (const char value) {
    wrlock ();
    d_buffer.add (value);
    unlock ();
  }

  // write a character string in the buffer

  void InputOutput::write (const char* value) {
    wrlock ();
    d_buffer.add (value, String::length (value));
    unlock ();
  }

  // create a new input-output stream in a generic way

  Object* InputOutput::mknew (Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();
    // check for 0 argument
    if (argc == 0) return new InputOutput;
    // check for 1 argument
    if (argc == 1) {
      String sval = argv->getstring (0);
      return new InputOutput (sval);
    }
    throw Exception ("argument-error", 
		     "invalid arguments with with input-output stream");
  }

  // apply this input-output stream with a set of arguments and a quark

  Object* InputOutput::apply (Runnable* robj, Nameset* nset, const long quark,
			      Vector* argv) {
    // check for input stream quark
    if (quark == QUARK_EOFP)    return Input::apply (robj, nset, quark, argv);
    if (quark == QUARK_READ)    return Input::apply (robj, nset, quark, argv);
    if (quark == QUARK_PUSHB)   return Input::apply (robj, nset, quark, argv);
    if (quark == QUARK_VALIDP)  return Input::apply (robj, nset, quark, argv);
    if (quark == QUARK_BUFLEN)  return Input::apply (robj, nset, quark, argv);

    // check for output stream quark
    if (quark == QUARK_WRITE)   return Input::apply (robj, nset, quark, argv);
    if (quark == QUARK_WRITELN) return Input::apply (robj, nset, quark, argv);
    if (quark == QUARK_ERRORLN) return Input::apply (robj, nset, quark, argv);
    if (quark == QUARK_NEWLINE) return Input::apply (robj, nset, quark, argv);

    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_RESET) {
	reset ();
	return nilp;
      }
    }

    // call the object method
    return Object::apply (robj, nset, quark, argv);
  }
}
