// ---------------------------------------------------------------------------
// - Set.cpp                                                                 -
// - standard object library - set class implementation                      -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2005 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Set.hpp"
#include "Input.hpp"
#include "Vector.hpp"
#include "Output.hpp"
#include "Boolean.hpp"
#include "Integer.hpp"
#include "Runnable.hpp"
#include "Exception.hpp"

namespace afnix {
 
  // supported set quarks
  static const long QUARK_ADD      = String::intern ("add");
  static const long QUARK_GETIT    = String::intern ("get-iterator");
  static const long QUARK_CARDINAL = String::intern ("cardinal");
  
  // -------------------------------------------------------------------------
  // - set element structure                                                 -
  // -------------------------------------------------------------------------

  struct s_selm {
    // the element object
    Object* p_obj;
    // the next element
    s_selm* p_nxt;
    
    // create a default element
    s_selm (void) {
      p_obj = nilp;
      p_nxt = nilp;
    }
    // create an element with an object
    s_selm (Object* obj) {
      if (obj == nilp) 
	throw Exception ("set-error", "cannot add nil object");
      Object::iref (p_obj = obj);
      p_nxt = nilp;
    }
    // make this element shared
    void mksho (void) {
      if (p_obj != nilp) p_obj->mksho ();
      if (p_nxt != nilp) p_nxt->mksho ();
    }
    // check if an object matches an element
    bool isobj (Object* obj) const {
      // check object and nil next
      if (p_obj == obj)  return true;
      if (p_nxt == nilp) return false;
      // go to next element
      return p_nxt->isobj (obj);
    }
  };

  // -------------------------------------------------------------------------
  // - set class section                                                     -
  // -------------------------------------------------------------------------

  // create a new empty set

  Set::Set (void) {
    d_size = 0;
    p_elem = nilp;
  }

  // create a new set with an object

  Set::Set (Object* obj) {
    d_size = 0;
    p_elem = nilp;
    add (obj);
  }

  // destroy this set

  Set::~Set (void) {
    delete p_elem;
  }

  // return the class name
  String Set::repr (void) const {
    return "Set";
  }

  // make this set a shared object

  void Set::mksho (void) {
    if (p_shared != nilp) return;
    Object::mksho ();
    if (p_elem != nilp) p_elem->mksho ();
  }

  // return the set serial code

  t_byte Set::serialid (void) const {
    return SERIAL_OSET_ID;
  }

  // serialize this set

  void Set::wrstream (Output& os) const {
    rdlock ();
    // write the set size
    Integer size (d_size);
    size.wrstream (os);
    // write the objects
    s_selm* elem = p_elem;
    while (elem != nilp) {
      Object* obj = elem->p_obj;
      if (obj == nilp) {
	unlock ();
	throw Exception ("serialize-error", 
			 "invalid nil object for set serialization");
      }
      // get the object serial form
      Serial* sobj = dynamic_cast <Serial*> (obj);
      if (sobj == nilp) {
	unlock ();
	throw Exception ("serial-error", "cannot serialize object", 
			 obj->repr ());
      }
      sobj->serialize (os);
      // get the next object
      elem = elem->p_nxt;
    }
    unlock ();
  }

  // deserialize this set

  void Set::rdstream (Input& is) {
    wrlock ();
    reset ();
    // get the vector length
    Integer slen;
    slen.rdstream (is);
    long size = slen.tointeger ();
    // read in each object
    for (long i = 0; i < size; i++) add (Serial::deserialize (is));
    unlock ();
  }

  // reset this set

  void Set::reset (void) {
    wrlock ();
    delete p_elem;
    p_elem = nilp;
    unlock ();
  }

  // return the cardinal of this set

  long Set::cardinal (void) const {
    rdlock ();
    long result = d_size;
    unlock ();
    return result;
  }

  // check if an object exists

  bool Set::exists (Object* obj) const {
    // do not take nil object
    if (obj == nilp) return false;
    rdlock ();
    // check that the object is not already there
    if ((p_elem != nilp) && (p_elem->isobj (obj) == true)) {
      unlock ();
      return true;
    }
    unlock ();
    return false;
  }

  // add an object to this set

  void Set::add (Object* obj) {
    // do not take nil object
    if (obj == nilp) return;
    wrlock ();
    // check that the object is not already there
    if ((p_elem != nilp) && (p_elem->isobj (obj) == true)) {
      unlock ();
      return;
    }
    try {
      // create a new element and link it
      s_selm* elem = new s_selm (obj);
      elem->p_nxt  = p_elem;
      p_elem = elem;
      // update the count
      d_size++;
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // return a set iterator

  Iterator* Set::makeit (void) {
    rdlock ();
    try {
      Setit* it = new Setit (this);
      unlock ();
      return it;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // create a new set in a generic way

  Object* Set::mknew (Vector* argv) {
    // create the set
    Set* result = new Set;
    // get the objects
    long argc = (argv == nilp) ? 0 : argv->length ();
    for (long i = 0; i < argc; i++) result->add (argv->get (i));
    return result;
  }

  // apply a set method by quark

  Object* Set::apply (Runnable* robj, Nameset* nset, const long quark,
		      Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_GETIT)    return makeit ();
      if (quark == QUARK_CARDINAL) return new Integer (cardinal ());
    }

    // dispatch 1 argument
    if (argc == 1) {
      if (quark == QUARK_ADD) {
	Object* result = argv->get (0);
	add (result);
	robj->post (result);
	return result;
      }
    }
    
    // call the object method
    return Object::apply (robj, nset, quark, argv);
  }

  // -------------------------------------------------------------------------
  // - set iterator class section                                            -
  // -------------------------------------------------------------------------

  // create a new set iterator

  Setit::Setit (Set* set) {
    Object::iref (p_set = set);
    if (p_set != nilp) p_set->rdlock ();
    p_elem = nilp;
    begin ();
  }

  // destroy this set iterator

  Setit::~Setit (void) {
    if (p_set != nilp) p_set->unlock ();
    Object::dref (p_set);
  }

  // return the class name

  String Setit::repr (void) const {
    return "Setit";
  }

  // make this set iterator a shared object

  void Setit::mksho (void) {
    if (p_shared != nilp) return;
    Object::mksho ();
    if (p_set != nilp) p_set->mksho ();
  }

  // reset the iterator to the begining

  void Setit::begin (void) {
    wrlock ();
    if (p_set != nilp) {
      p_elem = p_set->p_elem;
    }
    unlock ();
  }

  // reset the iterator to the end

  void Setit::end (void) {
    throw Exception ("iterator-error", "cannot set a set iterator to end");
  }

  // go to the next object

  void Setit::next (void) {
    wrlock ();
    if (p_elem != nilp) {
      p_elem = p_elem->p_nxt;
    }
    unlock ();
  }

  // go to the previous object
  void Setit::prev (void) {
    throw Exception ("iterator-error", "cannot move back a set iterator");
  }

  // get the object at the current position

  Object* Setit::getobj (void) {
    rdlock ();
    if (p_elem != nilp) {
      Object* result = p_elem->p_obj;
      unlock ();
      return result;
    }
    unlock ();
    return nilp;
  }

  // return true if the iterator is at the end

  bool Setit::isend (void) {
    rdlock ();
    bool result = (p_elem == nilp);
    unlock ();
    return result;
  }
}
