// ---------------------------------------------------------------------------
// - Folio.cpp                                                               -
// - afnix:sps library - folio class implementation                          -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2005 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Folio.hpp"
#include "Integer.hpp"
#include "Runnable.hpp"
#include "Exception.hpp"

namespace afnix {
  // the folio supported quarks
  static const long QUARK_ADD     = String::intern ("add");
  static const long QUARK_GET     = String::intern ("get");
  static const long QUARK_SET     = String::intern ("set");
  static const long QUARK_LENGTH  = String::intern ("length");
  static const long QUARK_GETNAME = String::intern ("get-name");
  static const long QUARK_SETNAME = String::intern ("set-name");

  // the folio magic number
  const long SPS_MSIZE   = 4;
  const char SPS_MAGIC[] = {'\377', 'O', 'D', 'B'};

  // this function write the folio header
  static void write_folio_magic (Output& os) {
    for (long i = 0; i < SPS_MSIZE; i++) os.write (SPS_MAGIC[i]);
  }

  // this function check that the header matches the edc magic number
  static bool check_folio_magic (Input* is) {
    // check for nil and reset
    if (is == nilp) return false;
    // read in the magic number
    char mbuf[SPS_MSIZE];
    for (long i = 0; i < SPS_MSIZE; i++) {
      mbuf[i] = is->read ();
      if (mbuf[i] != SPS_MAGIC[i]) {
	is->pushback (mbuf, i+1);
	return false;
      }
    }
    return true;
  }

  // create a nil folio

  Folio::Folio (void) {
    d_name = "";
    d_vsht.reset ();
  }

  // create a new folio by name

  Folio::Folio (const String& name) {
    d_name = name;
    d_vsht.reset ();
  }

  // create a new folio by stream

  Folio::Folio (Input* is) {
    // reset the folio
    d_name = "";
    d_vsht.reset ();
    // check for valid header
    if (check_folio_magic (is) == false)
      throw Exception ("folio-error", "invalid stream header");
    // read-in data
    rdstream (*is);
  }

  // return the object name

  String Folio::repr (void) const {
    return "Folio";
  }

  // write the folio to an output stream
  
  void Folio::write (Output& os) {
    rdlock ();
    try {
      // write the header
      write_folio_magic (os);
      // serialize the folio
      wrstream (os);
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // serialize a folio

  void Folio::wrstream (Output& os) const {
    rdlock ();
    // save the folio name
    d_name.wrstream (os);
    // save the vector
    d_vsht.wrstream (os);
    unlock ();
  }

  // deserialize this folio

  void Folio::rdstream (Input& is) {
    wrlock ();
    // get the folio name
    d_name.rdstream (is);
    // get the vector
    d_vsht.rdstream (is);
    unlock ();
  }

  // return the folio name

  String Folio::getname (void) const {
    rdlock ();
    String result = d_name;
    unlock ();
    return result;
  }

  // set the folio name

  void Folio::setname (const String& name) {
    wrlock ();
    d_name = name;
    unlock ();
  }

  // return the length of the folio

  long Folio::length (void) const {
    rdlock ();
    long result = d_vsht.length ();
    unlock ();
    return result;
  }

  // add a sheet in this folio

  void Folio::add (Sheet* sheet) {
    wrlock ();
    d_vsht.append (sheet);
    unlock ();
  }

  // get a folio sheet by index

  Sheet* Folio::get (const long index) const {
    rdlock ();
    try {
      Sheet* result = dynamic_cast <Sheet*> (d_vsht.get (index));
      unlock ();
      return result;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // set this folio by index and sheet

  void Folio::set (const long index, Sheet* sheet) {
    wrlock ();
    d_vsht.set (index, sheet);
    unlock ();
  }

  // create a new folio in a generic way

  Object* Folio::mknew (Vector* argv) {
    // get number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();
    // check for 0 argument
    if (argc == 0) return new Folio;
    // check for 1 argument
    if (argc == 1) {
      Object*  obj = argv->get (0);
      // check for a name
      String* sobj = dynamic_cast <String*> (obj);
      if (sobj != nilp) return new Folio (*sobj);
      // check for an input stream
      Input* iobj = dynamic_cast <Input*> (obj);
      if (iobj != nilp) return new Folio (iobj);
      // invalid argument
      throw Exception ("argument-error", "invalid folio argument",
		       Object::repr (obj));
    }
    throw Exception ("argument-error", "too many argument with folio");
  }

  // apply a folio method with a set of arguments and a quark

  Object* Folio::apply (Runnable* robj, Nameset* nset, const long quark,
			   Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_LENGTH ) return new Integer (length ());
      if (quark == QUARK_GETNAME) return new String  (getname ());
    }
    // dispatch 1 argument
    if (argc == 1) {
      if (quark == QUARK_SETNAME) {
	String name = argv->getstring (0);
	setname (name);
	return nilp;
      }
      if (quark == QUARK_GET) {
	long idx = argv->getint (0);
	rdlock ();
	try {
	  Object* result = get (idx);
	  robj->post (result);
	  unlock ();
	  return result;
	} catch (...) {
	  unlock ();
	  throw;
	}
      }
      if (quark == QUARK_ADD) {
	Object* obj = argv->get (0);
	Sheet*  sht = dynamic_cast <Sheet*> (obj);
	if ((obj != nilp) && (sht == nilp)) 
	  throw Exception ("type-error", "invalid object to add in folio",
			   obj->repr ());
	add (sht);
	return nilp;
      }
    }
    // dispatch 2 argument
    if (argc == 2) {
      if (quark == QUARK_SET) {
	long    idx = argv->getint (0);
	Object* obj = argv->get (1);
	Sheet*  sht = dynamic_cast <Sheet*> (obj);
	if ((obj != nilp) && (sht == nilp)) 
	  throw Exception ("type-error", "invalid object to set in folio",
			   obj->repr ());
	set (idx, sht);
	return nilp;
      }
    }
    // call the persist method
    return Persist::apply (robj, nset, quark, argv);
  }
}
