// ---------------------------------------------------------------------------
// - Importer.cpp                                                            -
// - afnix:sps library - importer class implementation                       -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2005 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Reader.hpp"
#include "Importer.hpp"
#include "Runnable.hpp"

namespace afnix {

  // the common supported quarks
  static const long QUARK_IMPORT = String::intern ("import");

  // -------------------------------------------------------------------------
  // - record importer class object                                          -
  // -------------------------------------------------------------------------

  // the record importer supported quarks
  static const long QUARK_SETRCD = String::intern ("set-record");
  static const long QUARK_GETRCD = String::intern ("get-record");

  // create a default record importer

  RecordImport::RecordImport (void) {
    p_ircd = nilp;
    setrcd (new Record);
  }

  // create a record importer with a record

  RecordImport::RecordImport (Record* rcd) {
    p_ircd = nilp;
    setrcd (rcd);
  }

  // destroy this record importer

  RecordImport::~RecordImport (void) {
    Object::dref (p_ircd);
  }

  // return the object name

  String RecordImport::repr (void) const {
    return "RecordImport";
  }

  // return the importation record

  Record* RecordImport::getrcd (void) const {
    rdlock ();
    Record* result = p_ircd;
    unlock ();
    return result;
  }

  // set the record to import

  void RecordImport::setrcd (Record* rcd) {
    wrlock ();
    Object::dref (p_ircd);
    Object::iref (p_ircd = rcd);
    unlock ();
  }

  // import a list of literal into the record

  void RecordImport::import (Cons* cons) {
    if (cons == nilp) return;
    wrlock ();
    try {
      // iterate in the cons cell
      while (cons != nilp) {
	// get the object and reduce it
	Object* car = cons->getcar ();
	Object* obj = (car == nilp) ? nilp : car->mini ();
	p_ircd->add (obj);
	cons = cons->getcdr ();
      }
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // import stream data into the record

  void RecordImport::import (Input* is) {
    if (is == nilp) return;
    wrlock ();
    // create a new reader
    Reader rd (is);
    wrlock ();
    Form* form = nilp;
    try {
      // parse with the reader
      while (true) {
	form = rd.parse ();
	if (form == nilp) break;
	import (form);
      }
    } catch (Exception& e) {
      if (form == nilp) {
	e.setlnum (rd.getlnum ());
      } else {
	e.setlnum (form->getlnum ());
	Object::cref (form);
      }
      unlock ();
      throw;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // create a new record importer in a generic way

  Object* RecordImport::mknew (Vector* argv) {
    // get number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();
    // check for 0 argument
    if (argc == 0) return new RecordImport;
    // check for 1 argument
    if (argc == 1) {
      Object* obj = argv->get (0);
      Record* rcd = dynamic_cast <Record*> (obj);
      if (rcd != nilp) {
	return new RecordImport (rcd);
      }
      throw Exception ("type-error", "invalid argument with record import");
    }
    throw Exception ("argument-error", "too many argument with record import");
  }

  // apply a record importer method with a set of arguments and a quark

  Object* RecordImport::apply (Runnable* robj, Nameset* nset, 
			       const long quark, Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_GETRCD) {
	rdlock ();
	Record* rcd = getrcd ();
	robj->post (rcd);
	unlock ();
	return rcd;
      }
    }
    // dispatch 1 argument
    if (argc == 1) {
      if (quark == QUARK_SETRCD) {
	Object* obj = argv->get (0);
	Record* rcd = dynamic_cast <Record*> (obj);
	if (rcd == nilp) 
	  throw Exception ("type-error", "invalid object for set-record ",
			   Object::repr (obj));
	setrcd (rcd);
	return nilp;
      }
      if (quark == QUARK_IMPORT) {
	Object* obj = argv->get (0);
	// check for an input stream
	Input*   is = dynamic_cast <Input*> (obj);
	if (is != nilp) {
	  import (is);
	  return nilp;
	}
	// check for a cons cell
	Cons* cons = dynamic_cast <Cons*> (obj);
	if (cons != nilp) {
	  import (cons);
	  return nilp;
	}
	throw Exception ("type-error", "invalid object for import ",
			 Object::repr (obj));
      }
    }
    // call the object method
    return Object::apply (robj, nset, quark, argv);
  }

  // -------------------------------------------------------------------------
  // - sheet importer class object                                           -
  // -------------------------------------------------------------------------

  // the sheet importer supported quarks
  static const long QUARK_SETSHT = String::intern ("set-sheet");
  static const long QUARK_GETSHT = String::intern ("get-sheet");

  // create a default sheet importer

  SheetImport::SheetImport (void) {
    p_isht = nilp;
    setsht (new Sheet);
  }

  // create a sheet importer with a sheet

  SheetImport::SheetImport (Sheet* rcd) {
    p_isht = nilp;
    setsht (rcd);
  }

  // destroy this sheet importer

  SheetImport::~SheetImport (void) {
    Object::dref (p_isht);
  }

  // return the object name

  String SheetImport::repr (void) const {
    return "SheetImport";
  }

  // return the importation sheet

  Sheet* SheetImport::getsht (void) const {
    rdlock ();
    Sheet* result = p_isht;
    unlock ();
    return result;
  }

  // set the sheet to import

  void SheetImport::setsht (Sheet* sht) {
    wrlock ();
    Object::dref (p_isht);
    Object::iref (p_isht = sht);
    unlock ();
  }

  // import a list of literal into the sheet

  void SheetImport::import (Cons* cons) {
    if (cons == nilp) return;
    wrlock ();
    Record* rcd = new Record;
    try {
      // iterate in the cons cell
      while (cons != nilp) {
	// get the object and reduce it
	Object* car = cons->getcar ();
	Object* obj = (car == nilp) ? nilp : car->mini ();
	rcd->add (obj);
	cons = cons->getcdr ();
      }
      p_isht->add (rcd);
    } catch (...) {
      Object::cref (rcd);
      unlock ();
      throw;
    }
  }

  // import stream data into the sheet

  void SheetImport::import (Input* is) {
    if (is == nilp) return;
    wrlock ();
    // create a new reader
    Reader rd (is);
    wrlock ();
    Form* form = nilp;
    try {
      // parse with the reader
      while (true) {
	form = rd.parse ();
	if (form == nilp) break;
	import (form);
      }
    } catch (Exception& e) {
      if (form == nilp) {
	e.setlnum (rd.getlnum ());
      } else {
	e.setlnum (form->getlnum ());
	Object::cref (form);
      }
      unlock ();
      throw;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // create a new sheet importer in a generic way

  Object* SheetImport::mknew (Vector* argv) {
    // get number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();
    // check for 0 argument
    if (argc == 0) return new SheetImport;
    // check for 1 argument
    if (argc == 1) {
      Object* obj = argv->get (0);
      Sheet*  sht = dynamic_cast <Sheet*> (obj);
      if (sht != nilp) {
	return new SheetImport (sht);
      }
      throw Exception ("type-error", "invalid argument with sheet import");
    }
    throw Exception ("argument-error", "too many argument with sheet import");
  }

  // apply a sheet importer method with a set of arguments and a quark

  Object* SheetImport::apply (Runnable* robj, Nameset* nset, 
			      const long quark, Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_GETSHT) {
	rdlock ();
	Sheet* sht = getsht ();
	robj->post (sht);
	unlock ();
	return sht;
      }
    }
    // dispatch 1 argument
    if (argc == 1) {
      if (quark == QUARK_SETSHT) {
	Object* obj = argv->get (0);
	Sheet*  sht = dynamic_cast <Sheet*> (obj);
	if (sht == nilp) 
	  throw Exception ("type-error", "invalid object for set-sheet ",
			   Object::repr (obj));
	setsht (sht);
	return nilp;
      }
      if (quark == QUARK_IMPORT) {
	Object* obj = argv->get (0);
	// check for an input stream
	Input*   is = dynamic_cast <Input*> (obj);
	if (is != nilp) {
	  import (is);
	  return nilp;
	}
	// check for a cons cell
	Cons* cons = dynamic_cast <Cons*> (obj);
	if (cons != nilp) {
	  import (cons);
	  return nilp;
	}
	throw Exception ("type-error", "invalid object for import ",
			 Object::repr (obj));
      }
    }
    // call the object method
    return Object::apply (robj, nset, quark, argv);
  }
}
