// ---------------------------------------------------------------------------
// - Sheet.cpp                                                               -
// - afnix:sps library - sheet class implementation                          -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2005 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Sheet.hpp"
#include "Input.hpp"
#include "Spssid.hpp"
#include "Boolean.hpp"
#include "Integer.hpp"
#include "Importer.hpp"
#include "Runnable.hpp"
#include "Exception.hpp"

namespace afnix {

  // -------------------------------------------------------------------------
  // - sheet sort class object                                               -
  // -------------------------------------------------------------------------

  class SheetSorter {
  private:
    // the column index
    long d_col;
    // the sorting mode
    bool d_mode;

    // compare two objects for ascending order
    bool cmplth (Object* ref, Object* slv) const {
      // check that we have a object records
      Record* rref = dynamic_cast <Record*> (ref);
      Record* rslv = dynamic_cast <Record*> (slv);
      // map the cell to their object
      Object* robj = nilp;
      try {
	robj = (rref == nilp) ? nilp : rref->map (d_col);
      } catch (...) {
	robj = nilp;
      }
      Object* sobj = nilp;
      try {
	sobj = (rslv == nilp) ? nilp : rslv->map (d_col);
      } catch (...) {
	sobj = nilp;
      }
      // compare with operator
      if (robj == nilp) return true;
      Object*   obj = robj->oper (Object::LTH, sobj);
      Boolean* bobj = dynamic_cast <Boolean*> (obj);
      bool result   = (bobj == nilp) ? false : bobj->toboolean ();
      Object::cref (obj);
      return result;
    }

    // compare two objects for descending order
    bool cmpgth (Object* ref, Object* slv) const {
      // check that we have a object records
      Record* rref = dynamic_cast <Record*> (ref);
      Record* rslv = dynamic_cast <Record*> (slv);
      // map the cell to their object
      Object* robj = nilp;
      try {
	robj = (rref == nilp) ? nilp : rref->map (d_col);
      } catch (...) {
	robj = nilp;
      }
      Object* sobj = nilp;
      try {
	sobj = (rslv == nilp) ? nilp : rslv->map (d_col);
      } catch (...) {
	sobj = nilp;
      }
      // compare with operator
      if ((robj == nilp) && (sobj == nilp)) return true;
      if ((robj == nilp) && (sobj != nilp)) return false;
      Object*   obj = robj->oper (Object::GTH, sobj);
      Boolean* bobj = dynamic_cast <Boolean*> (obj);
      bool result   = (bobj == nilp) ? false : bobj->toboolean ();
      Object::cref (obj);
      return result;
    }

    // partition the vector and returns the pivot index

    long partition (Vector& argv, const long first, const long last) const {
      // initialize the pivot index, last index and next (unknown) index
      long pidx = first;
      long lidx = first;
      long uidx = lidx + 1;
      // get the pivot object - and protect it
      Object* pvt = Object::iref (argv.get (pidx));
      // swap until we have reached the last element
      while (uidx <= last) {
	Object* ref = Object::iref (argv.get (uidx));
	bool   cflg = d_mode ? cmplth (ref, pvt) : cmpgth (ref, pvt);
	if (cflg == true) {
	  lidx++;
	  argv.set (uidx, argv.get (lidx));
	  argv.set (lidx, ref);
	}
	Object::tref (ref);
	// the intel is in the other region
	uidx++;
      }
      // place the pivot in proper position
      argv.set (first, argv.get (lidx));
      argv.set (lidx,  pvt);
      Object::tref (pvt);
      // the pivot index is now the last index
      return lidx;
    }
    
    // this function performs a recursive quick sort

    void quicksort (Vector& argv, const long first, const long last) const {
      if (first >= last) return;
      long pidx = partition (argv, first, last);
      quicksort (argv, first, pidx - 1);
      quicksort (argv, pidx + 1, last);
    }

  public:
    // create a default sheet sorter
    SheetSorter (void) {
      d_col  = 0;
      d_mode = false;
    }

    // create a sorting object by index
    SheetSorter (const long col, const bool mode) {
      d_col  = col;
      d_mode = mode;
    }

    // sort a vector

    void qsort (Vector& argv) const {
      // sort the vector in place
      argv.wrlock ();
      try {
	long argc = argv.length ();
	quicksort (argv, 0, argc - 1);
	argv.unlock ();
      } catch (...) {
	argv.unlock ();
	throw;
      }
    }
  };

  // -------------------------------------------------------------------------
  // - sheet class object                                                    -
  // -------------------------------------------------------------------------

  // the sheet supported quarks
  static const long QUARK_ADD       = String::intern ("add");
  static const long QUARK_GET       = String::intern ("get");
  static const long QUARK_SET       = String::intern ("set");
  static const long QUARK_SORT      = String::intern ("sort");
  static const long QUARK_IMPORT    = String::intern ("import");
  static const long QUARK_LENGTH    = String::intern ("length");
  static const long QUARK_CONVERT   = String::intern ("convert");
  static const long QUARK_GETNAME   = String::intern ("get-name");
  static const long QUARK_SETNAME   = String::intern ("set-name");
  static const long QUARK_ADDDATA   = String::intern ("add-data");
  static const long QUARK_ADDINFO   = String::intern ("add-info");
  static const long QUARK_GETINFO   = String::intern ("get-info");
  static const long QUARK_MAPINFO   = String::intern ("map-info");
  static const long QUARK_SETINFO   = String::intern ("set-info");
  static const long QUARK_ADDHEAD   = String::intern ("add-header");
  static const long QUARK_GETHEAD   = String::intern ("get-header");
  static const long QUARK_MAPHEAD   = String::intern ("map-header");
  static const long QUARK_SETHEAD   = String::intern ("set-header");
  static const long QUARK_ADDFOOT   = String::intern ("add-footer");
  static const long QUARK_GETFOOT   = String::intern ("get-footer");
  static const long QUARK_MAPFOOT   = String::intern ("map-footer");
  static const long QUARK_SETFOOT   = String::intern ("set-footer");

  // this function computes the maximum between two numbers
  static inline long max (const long x, const long y) {
    return (x < y) ? y : x;
  }

  // this procedure returns a new sheet object for deserialization
  static Serial* mksob (void) {
    return new Sheet;
  }
  // register this cell serial id
  static const t_byte SERIAL_ID = Serial::setsid (SERIAL_SHTT_ID, mksob);

  // create a nil sheet

  Sheet::Sheet (void) {
    d_quark = 0;
  }

  // create a new sheet by name

  Sheet::Sheet (const String& name) {
    d_quark = name.toquark ();
  }

  // return the object name

  String Sheet::repr (void) const {
    return "Sheet";
  }

  // return the sheet serial id

  t_byte Sheet::serialid (void) const {
    return SERIAL_SHTT_ID;
  }

  // serialize a sheet

  void Sheet::wrstream (Output& os) const {
    rdlock ();
    // save the sheet name
    const String& name = String::qmap (d_quark);
    name.wrstream (os);
    // save the descriptor
    d_info.wrstream (os);
    // save the header
    d_head.wrstream (os);
    // save the footer
    d_foot.wrstream (os);
    // save the vector
    d_vrcds.wrstream (os);
    unlock ();
  }

  // deserialize this sheet

  void Sheet::rdstream (Input& is) {
    wrlock ();
    // get the sheet name
    String name;
    name.rdstream (is);
    d_quark = name.toquark ();
    // get the descriptor
    d_info.rdstream (is);
    // get the header
    d_head.rdstream (is);
    // get the footer
    d_foot.rdstream (is);
    // get the vector
    d_vrcds.rdstream (is);
    unlock ();
  }

  // return the sheet name

  String Sheet::getname (void) const {
    rdlock ();
    String result = String::qmap (d_quark);
    unlock ();
    return result;
  }

  // set the sheet name

  void Sheet::setname (const String& name) {
    rdlock ();
    d_quark = name.toquark ();
    unlock ();
  }

  // add a literal in the info descriptor
  
  void Sheet::addinfo (Object* object) {
    d_info.add (object);
  }

  // add a vector of objects in the info record

  void Sheet::addinfo (const Vector* argv) {
    long argc = (argv == nilp) ? 0 : argv->length ();
    if (argc == 0) return;
    wrlock ();
    try {
      for (long i = 0; i < argc; i++) d_info.add (argv->get (i));
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // get an info cell by index

  Cell* Sheet::getinfo (const long index) const {
    rdlock ();
    try {
      Cell* result = dynamic_cast <Cell*> (d_info.get (index));
      unlock ();
      return result;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // evaluate an info cell by index

  Literal* Sheet::mapinfo (const long index) const {
    rdlock ();
    try {
      Cell* cell = getinfo (index);
      Literal* result = (cell == nilp) ? nilp : cell->get ();
      unlock ();
      return result;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // set the information record by index and literal
  
  void Sheet::setinfo (const long index, Literal* lobj) {
    wrlock ();
    try {
      Cell* cell = dynamic_cast <Cell*> (d_info.get (index)); 
      if (cell != nilp) cell->set (lobj);
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // add an object in the header record
  
  void Sheet::addhead (Object* object) {
    d_head.add (object);
  }

  // add a vector of objects in the header record

  void Sheet::addhead (const Vector* argv) {
    long argc = (argv == nilp) ? 0 : argv->length ();
    if (argc == 0) return;
    wrlock ();
    try {
      for (long i = 0; i < argc; i++) d_head.add (argv->get (i));
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // get a header cell by index

  Cell* Sheet::gethead (const long index) const {
    rdlock ();
    try {
      Cell* result = dynamic_cast <Cell*> (d_head.get (index));
      unlock ();
      return result;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // evaluate a header cell by index

  Literal* Sheet::maphead (const long index) const {
    rdlock ();
    try {
      Cell* cell = gethead (index);
      Literal* result = (cell == nilp) ? nilp : cell->get ();
      unlock ();
      return result;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // set a record header by index and literal
  
  void Sheet::sethead (const long index, Literal* lobj) {
    wrlock ();
    try {
      Cell* cell = dynamic_cast <Cell*> (d_head.get (index)); 
      if (cell != nilp) cell->set (lobj);
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // add an object in the footer record
  
  void Sheet::addfoot (Object* object) {
    d_foot.add (object);
  }

  // add a vector of objects in the footer record

  void Sheet::addfoot (const Vector* argv) {
    long argc = (argv == nilp) ? 0 : argv->length ();
    if (argc == 0) return;
    wrlock ();
    try {
      for (long i = 0; i < argc; i++) d_foot.add (argv->get (i));
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // get a footer cell by index

  Cell* Sheet::getfoot (const long index) const {
    rdlock ();
    try {
      Cell* result = dynamic_cast <Cell*> (d_foot.get (index));
      unlock ();
      return result;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // evaluate a footer cell by index

  Literal* Sheet::mapfoot (const long index) const {
    rdlock ();
    try {
      Cell* cell = getfoot (index);
      Literal* result = (cell == nilp) ? nilp : cell->get ();
      unlock ();
      return result;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // set a record footer by index and literal
  
  void Sheet::setfoot (const long index, Literal* lobj) {
    wrlock ();
    try {
      Cell* cell = dynamic_cast <Cell*> (d_foot.get (index)); 
      if (cell != nilp) cell->set (lobj);
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // add a record in this sheet

  void Sheet::add (Record* rcd) {
    if (rcd == nilp) return;
    wrlock ();
    d_vrcds.append (rcd);
    unlock ();
  }

  // get a record by index

  Record* Sheet::get (const long index) const {
    rdlock ();
    Record* result = dynamic_cast <Record*> (d_vrcds.get (index));
    unlock ();
    return result;
  }

  // set a record in this sheet by index

  void Sheet::set (const long index, Record* rcd) {
    wrlock ();
    d_vrcds.set (index, rcd);
    unlock ();
  }

  // add a vector of literal as a record

  void Sheet::adddata (const Vector* argv) {
    long argc = (argv == nilp) ? 0 : argv->length ();
    if (argc == 0) return;
    wrlock ();
    Record* rcd = new Record;
    try {
      for (long i = 0; i < argc; i++) rcd->add (argv->get (i));
      d_vrcds.append (rcd);
      unlock ();
    } catch (...) {
      Object::cref (rcd);
      unlock ();
      throw;
    }
  }

  // return the length of the sheet list

  long Sheet::length (void) const {
    rdlock ();
    long result = d_vrcds.length ();
    unlock ();
    return result;
  }

  // return the number of columns
  
  long Sheet::getcols (void) const {
    rdlock ();
    // get the sheet length
    long tlen = length ();
    // compute the maximum columns
    long result = 0;
    for (long i = 0; i < tlen; i++) {
      Record* rcd = get (i);
      if (rcd == nilp) continue;
      result = max (result, rcd->length ());
    }
    unlock ();
    return result;
  }

  // import data in this sheet

  void Sheet::import (Input* is) {
    if (is == nilp) return;
    // protect us during the importation
    Object::iref (this);
    // create a new importer
    SheetImport ti (this);
    try {
      ti.import (is);
      Object::tref (this);
    } catch (...) {
      Object::tref (this);
      throw;
    }
  }

  // convert this sheet into a print sheet

  PrintTable* Sheet::convert (long max, long start, bool flag) const {
    rdlock ();
    // get the number of rows
    long rows = length ();
    // check for start index
    if ((start < 0) || (start >= rows)) {
      unlock ();
      throw Exception ("sheet-error", "start index out of range for convert");
    }
    // check for max index
    long tlen = start + ((max == 0) ? rows : max);
    if (tlen > rows) {
      unlock ();
      throw Exception ("sheet-error", "max index is out of range for convert");
    }
    // format result
    long cols = getcols ();
    PrintTable* result = new PrintTable (cols);
    try {
      // iterate through the record
      for (long i = start; i < tlen; i++) {
	Record* rcd = get (i);
	if (rcd == nilp) continue;
	long row = result->add ();
	// get the record length
	long rlen = rcd->length ();
	for (long j = 0; j < rlen; j++) {
	  Literal* lobj = rcd->map (j);
	  if (lobj == nilp) {
	    result->set (row, j, "nil");
	  } else {
	    String data = flag ? lobj->tostring () : lobj->toliteral ();
	    result->set (row, j, data);
	  }
	}
	// fill the rest of the sheet
	for (long j = rlen; j < cols; j++) result->set (row, j, "nil");
      }
    } catch (...) {
      delete result;
      unlock ();
      throw;
    }
    unlock ();
    return result;
  }

  // sort this sheet by column index and mode

  void Sheet::sort (const long col, const bool mode) {
    wrlock ();
    // build the sorter object
    SheetSorter sorter (col, mode);
    // sort the vector
    try {
      sorter.qsort (d_vrcds);
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // create a new sheet in a generic way

  Object* Sheet::mknew (Vector* argv) {
    // get number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();
    // check for 0 argument
    if (argc == 0) return new Sheet;
    // check for 1 argument
    if (argc == 1) {
      String name = argv->getstring (0);
      return new Sheet (name);
    }
    throw Exception ("argument-error", "too many argument with sheet");
  }

  // apply a sheet method with a set of arguments and a quark

  Object* Sheet::apply (Runnable* robj, Nameset* nset, const long quark,
			Vector* argv) {
    // check for generic quark
    if (quark == QUARK_ADDDATA) {
      adddata (argv);
      return nilp;
    }
    if (quark == QUARK_ADDINFO) {
      addinfo (argv);
      return nilp;
    }
    if (quark == QUARK_ADDHEAD) {
      addhead (argv);
      return nilp;
    }
    if (quark == QUARK_ADDFOOT) {
      addfoot (argv);
      return nilp;
    }

    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_LENGTH ) return new Integer (length  ());
      if (quark == QUARK_GETNAME) return new String  (getname ());
      if (quark == QUARK_CONVERT) return convert (0, 0, true);
      if (quark == QUARK_SORT) {
	sort (0, true);
	return nilp;
      }
    }

    // dispatch 1 argument
    if (argc == 1) {
      if (quark == QUARK_SETNAME) {
	String name = argv->getstring (0);
	setname (name);
	return nilp;
      }
      if (quark == QUARK_GET) {
	long idx = argv->getint (0);
	rdlock ();
	try {
	  Object* result = get (idx);
	  robj->post (result);
	  unlock ();
	  return result;
	} catch (...) {
	  unlock ();
	  throw;
	}
      }
      if (quark == QUARK_ADD) {
	Object* obj = argv->get (0);
	Record* rcd = dynamic_cast <Record*> (obj);
	if (rcd != nilp) {
	  add (rcd);
	  return nilp;
	}
	throw Exception ("type-error", "invalid object to add in sheet",
			 Object::repr (obj));
      }
      if (quark == QUARK_GETINFO) {
	long index = argv->getint (0);
	rdlock ();
	try {
	  Object* result = getinfo (index);
	  robj->post (result);
	  unlock ();
	  return result;
	} catch (...) {
	  unlock ();
	  throw;
	}
      }
      if (quark == QUARK_MAPINFO) {
	long index = argv->getint (0);
	rdlock ();
	try {
	  Object* result = mapinfo (index);
	  robj->post (result);
	  unlock ();
	  return result;
	} catch (...) {
	  unlock ();
	  throw;
	}
      }
      if (quark == QUARK_GETHEAD) {
	long index = argv->getint (0);
	rdlock ();
	try {
	  Object* result = gethead (index);
	  robj->post (result);
	  unlock ();
	  return result;
	} catch (...) {
	  unlock ();
	  throw;
	}
      }
      if (quark == QUARK_MAPHEAD) {
	long index = argv->getint (0);
	rdlock ();
	try {
	  Object* result = maphead (index);
	  robj->post (result);
	  unlock ();
	  return result;
	} catch (...) {
	  unlock ();
	  throw;
	}
      }
      if (quark == QUARK_GETFOOT) {
	long index = argv->getint (0);
	rdlock ();
	try {
	  Object* result = getfoot (index);
	  robj->post (result);
	  unlock ();
	  return result;
	} catch (...) {
	  unlock ();
	  throw;
	}
      }
      if (quark == QUARK_MAPFOOT) {
	long index = argv->getint (0);
	rdlock ();
	try {
	  Object* result = mapfoot (index);
	  robj->post (result);
	  unlock ();
	  return result;
	} catch (...) {
	  unlock ();
	  throw;
	}
      }
      if (quark == QUARK_IMPORT) {
	Object* obj = argv->get (0);
	Input*   is = dynamic_cast <Input*> (obj);
	if (is != nilp) {
	  import (is);
	  return nilp;
	}
	throw Exception ("type-error", "invalid object for sheet importation",
			 Object::repr (obj));
      }
      if (quark == QUARK_CONVERT) {
	long max = argv->getint (0);
	return convert (max, 0, true);
      }
      if (quark == QUARK_SORT) {
	Object* obj = argv->get (0);
	// check for integer
	Integer* col = dynamic_cast <Integer*> (obj);
	if (col != nilp) {
	  sort (col->tointeger (), true);
	  return nilp;
	}
	// check for boolean
	Boolean* mode = dynamic_cast <Boolean*> (obj);
	if (mode != nilp) {
	  sort (0, mode->toboolean ());
	  return nilp;
	}
	// invalid arguments
	throw Exception ("type-error", "invalid object for sorting ",
			 Object::repr (obj));
      }
    }

    // dispatch 2 argument
    if (argc == 2) {
      if (quark == QUARK_SETINFO) {
	long    idx = argv->getint (0);
	Object*  obj  = argv->get (1);
	Literal* lobj = dynamic_cast <Literal*> (obj);
	if ((obj != nilp) && (lobj == nilp)) 
	  throw Exception ("type-error", "invalid object to set in info",
			   obj->repr ());
	setinfo (idx, lobj);
	return nilp;
      }
      if (quark == QUARK_SETHEAD) {
	long    idx = argv->getint (0);
	Object*  obj  = argv->get (1);
	Literal* lobj = dynamic_cast <Literal*> (obj);
	if ((obj != nilp) && (lobj == nilp)) 
	  throw Exception ("type-error", "invalid object to set in header",
			   obj->repr ());
	sethead (idx, lobj);
	return nilp;
      }
      if (quark == QUARK_SETFOOT) {
	long    idx = argv->getint (0);
	Object*  obj  = argv->get (1);
	Literal* lobj = dynamic_cast <Literal*> (obj);
	if ((obj != nilp) && (lobj == nilp)) 
	  throw Exception ("type-error", "invalid object to set in footer",
			   obj->repr ());
	setfoot (idx, lobj);
	return nilp;
      }
      if (quark == QUARK_SET) {
	long    idx = argv->getint (0);
	Object* obj = argv->get (1);
	Record* rcd = dynamic_cast <Record*> (obj);
	if ((obj != nilp) && (rcd == nilp)) 
	  throw Exception ("type-error", "invalid object to set in sheet",
			   obj->repr ());
	set (idx, rcd);
	return nilp;
      }
      if (quark == QUARK_CONVERT) {
	long max   = argv->getint (0);
	long start = argv->getint (1);
	return convert (max, start, true);
      }
      if (quark == QUARK_SORT) {
	long col  = argv->getint  (0);
	bool mode = argv->getbool (1);
	sort (col, mode);
	return nilp;
      }
    }
    // dispatch 3 arguments
    if (argc == 3) {
      if (quark == QUARK_CONVERT) {
	long max   = argv->getint  (0);
	long start = argv->getint  (1);
	bool flag  = argv->getbool (2);
	return convert (max, start, flag);
      }
    }
    // call the persist method
    return Persist::apply (robj, nset, quark, argv);
  }
}
