// ---------------------------------------------------------------------------
// - Literate.cpp                                                            -
// - afnix:txt library - transliterate object class implementation           -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2005 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Vector.hpp"
#include "Integer.hpp"
#include "Literate.hpp"
#include "Character.hpp"
#include "Exception.hpp"
 
namespace afnix {

  // the literate supported quarks
  static const long QUARK_READ   = String::intern ("read");
  static const long QUARK_RESET  = String::intern ("reset");
  static const long QUARK_SETMAP = String::intern ("set-map");
  static const long QUARK_GETMAP = String::intern ("get-map");
  static const long QUARK_SETESC = String::intern ("set-escape");
  static const long QUARK_GETESC = String::intern ("get-escape");
  static const long QUARK_SETEMC = String::intern ("set-escape-map");
  static const long QUARK_GETEMC = String::intern ("get-escape-map");
  static const long QUARK_TRANSL = String::intern ("translate");

  // create a default literate

  Literate::Literate (void) {
    reset ();
  }

  // create a literate with a escape character

  Literate::Literate (const char escc) {
    reset ();
    d_escc = escc;
  }

  // create a literate with a model

  Literate::Literate (const t_model model) {
    setmodel (model);
  }

  // return the class name

  String Literate::repr (void) const {
    return "Literate";
  }

  // reset this literate

  void Literate::reset (void) {
    wrlock ();
    // reset mapping table
    for (long i = 0; i < 256; i++) d_mapt[i] = i;
    // reset the escape character and table
    d_escc = nilc;
    for (long i = 0; i < 256; i++) d_esct[i] = nilc;    
    unlock ();
  }

  // set a transliteration model

  void Literate::setmodel (const t_model model) {
    // always reset the model
    reset ();
    // the afnix escaped string model
    if (model == ASTR) {
      setesc ('\\');
      setemc ('t',  tabc);
      setemc ('n',  eolc);
      setemc ('\\', '\\');
    }
  }

  // set a mapping character

  void Literate::setmap (const char mapc, const char altc) {
    wrlock ();
    long cidx    = static_cast <long> (mapc);
    d_mapt[cidx] = altc;
    unlock ();
  }

  // return a mapping character

  char Literate::getmap (const char mapc) const {
    rdlock ();
    long cidx   = static_cast <long> (mapc);
    char result = d_mapt[cidx];
    unlock ();
    return result;
  }

  // set the escape character

  void Literate::setesc (const char escc) {
    wrlock ();
    d_escc = escc;
    unlock ();
  }

  // return the escape character

  char Literate::getesc (void) const {
    rdlock ();
    char result = d_escc;
    unlock ();
    return result;
  }

  // set an escaped mapping character

  void Literate::setemc (const char mapc, const char altc) {
    wrlock ();
    long cidx    = static_cast <long> (mapc);
    d_esct[cidx] = altc;
    unlock ();
  }

  // return an escaped mapping character

  char Literate::getemc (const char mapc) const {
    rdlock ();
    long cidx   = static_cast <long> (mapc);
    char result = d_esct[cidx];
    unlock ();
    return result;
  }

  // transliterate a string into another string

  String Literate::translate (const String& s) const {
    rdlock ();
    try {
      // create a result buffer
      Buffer buf;
      // loop in the string
      long len = s.length ();
      long max = len - 1;
      for (long i = 0; i < len; i++) {
	long cidx = static_cast <long> (s[i]);
	// map the character
	char altc = d_mapt[cidx];
	// check for escape
	if ((i < max) && (d_escc != nilc) && (altc == d_escc)) {
	  long nc = static_cast <long> (s[++i]);
	  char ec = d_esct[nc];
	  if (ec!= nilc) altc = ec;
	}
	// add it into the buffer
	buf.add (altc);
      }
      unlock ();
      return buf.tostring ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // read a character from an input stream and map it

  char Literate::read (Input& is) const {
    rdlock ();
    try {
      long cidx = static_cast <long> (is.read ());
      // map the character
      char altc = d_mapt[cidx];
      // check for escape
      if ((d_escc != nilc) && (altc == d_escc)) {
	char nc = is.read ();
	long ci = static_cast <long> (nc);
	char ec = d_esct[ci];
	if (ec!= nilc) 
	  altc = ec;
	else
	  is.pushback (nc);
      }
      unlock ();
      return altc;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // create a new literate in a generic way

  Object* Literate::mknew (Vector* argv) {
    long argc = (argv == nilp) ? 0 : argv->length ();
    // check for 0 argument
    if (argc == 0) return new Literate;
    // check for 1 argument
    if (argc == 1) {
      char escc = argv->getchar (0);
      return new Literate (escc);
    }
    throw Exception ("argument-error", "too many arguments with literate");
  }

  // apply this literate with a set of arguments and a quark

  Object* Literate::apply (Runnable* robj, Nameset* nset, const long quark,
			   Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_GETESC) return new Character (getesc ());
      if (quark == QUARK_RESET) {
	reset ();
	return nilp;
      }
    }
    // dispatch one argument
    if (argc == 1) {
      if (quark == QUARK_GETMAP) {
	char mapc = argv->getchar (0);
	return new Character (getmap (mapc));
      }
      if (quark == QUARK_SETESC) {
	char escc = argv->getchar (0);
	setesc (escc);
	return nilp;
      }
      if (quark == QUARK_GETEMC) {
	char mapc = argv->getchar (0);
	return new Character (getemc (mapc));
      }
      if (quark == QUARK_TRANSL) {
	String sval = argv->getstring (0);
	return new String (translate (sval));
      }
      if (quark == QUARK_READ) {
	Object* obj = argv->get (0);
	// check for an input stream
	Input* is = dynamic_cast <Input*> (obj);
	if (is == nilp) throw Exception ("type-error",
					 "invalid object with read method",
					 Object::repr (obj));
	return new Character (read (*is));
      }
    }
    // dispatch two arguments
    if (argc == 2) {
      if (quark == QUARK_SETMAP) {
	char mapc = argv->getchar (0);
	char altc = argv->getchar (1);
	setmap (mapc, altc);
	return nilp;
      }
      if (quark == QUARK_SETEMC) {
	char mapc = argv->getchar (0);
	char altc = argv->getchar (1);
	setemc (mapc, altc);
	return nilp;
      }
    }
    // call the object method
    return Object::apply (robj, nset, quark, argv);
  }
}
