// ---------------------------------------------------------------------------
// - HtmlDiv.cpp                                                             -
// - afnix:www library - html div element class implementation               -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2005 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "HtmlDiv.hpp"
#include "Runnable.hpp"
#include "Exception.hpp"

namespace afnix {

  // the element supported quarks
  static const long QUARK_GETLINK = String::intern ("get-link");
  static const long QUARK_SETLINK = String::intern ("set-link");

  // create a default div element

  HtmlDiv::HtmlDiv (void) : HtmlElement ("div") {
    p_link = nilp;
  }

  // create an element by name

  HtmlDiv::HtmlDiv (const String& cname) : HtmlElement ("div") {
    p_link = nilp;
    if (cname.length () != 0) addtag ("class", cname);
  }

  // destroy this html div element

  HtmlDiv::~HtmlDiv (void) {
    Object::dref (p_link);
  }

  // return the class name

  String HtmlDiv::repr (void) const {
    return "HtmlDiv";
  }

  // set the link element
  
  void HtmlDiv::setlink (HtmlElement* elem) {
    wrlock ();
    Object::iref (elem);
    Object::dref (p_link);
    p_link = elem;
    unlock ();
  }

  // return the link element

  HtmlElement* HtmlDiv::getlink (void) const {
    rdlock ();
    HtmlElement* elem = p_link;
    unlock ();
    return elem;
  }

  // emit this element into a buffer

  void HtmlDiv::emit (Buffer& buf) const {
    rdlock  ();
    buf.add (opening ());
    if (p_link != nilp) p_link->emit (buf);
    buf.add (closing ());
    buf.add (eolc);
    unlock ();
  }

  // create a new div element object in a generic way

  Object* HtmlDiv::mknew (Vector* argv) {
    long argc = (argv == nilp) ? 0 : argv->length ();

    // check for no argument
    if (argc == 0) return new HtmlDiv;
    // check for one argument
    if (argc == 1) {
      String cname = argv->getstring (0);
      return new HtmlDiv (cname);
    }
    // wrong arguments
    throw Exception ("argument-error", 
		     "invalid arguments with html div element object");
  }

  // apply this html div element object with a set of arguments and a quark

  Object* HtmlDiv::apply (Runnable* robj, Nameset* nset, const long quark,
			  Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_GETLINK) {
	rdlock ();
	try {
	  HtmlElement* elem = getlink ();
	  robj->post (elem);
	  unlock ();
	  return elem;
	} catch (...) {
	  unlock ();
	  throw;
	}
      }
    }

    // dispatch 1 argument
    if (argc == 1) {
      if (quark == QUARK_SETLINK) {
	HtmlElement* elem = dynamic_cast <HtmlElement*> (argv->get (0));
	if (elem == nilp) {
	  throw Exception ("type-error", "div element expected with set-link");
	}
	setlink (elem);
	return nilp;
      }
    }
    // call the element method
    return HtmlElement::apply (robj, nset, quark, argv);
  }
}
