// ---------------------------------------------------------------------------
// - HtmlElement.cpp                                                         -
// - afnix:www library - html element class implementation                   -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2005 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Exception.hpp"
#include "HtmlElement.hpp"

namespace afnix {

  // the element supported quarks
  static const long QUARK_EMIT    = String::intern ("emit");
  static const long QUARK_RESET   = String::intern ("reset");
  static const long QUARK_ADDTAG  = String::intern ("add-tag");
  static const long QUARK_GETNAME = String::intern ("get-name");
  static const long QUARK_OPENING = String::intern ("opening");
  static const long QUARK_CLOSING = String::intern ("closing");

  // create an element by name

  HtmlElement::HtmlElement (const String& name) {
    d_name = name;
    d_tags.reset ();
  }

  // return the class name

  String HtmlElement::repr (void) const {
    return "HtmlElement";
  }

  // reset this element

  void HtmlElement::reset (void) {
    wrlock ();
    d_tags.reset ();
    unlock ();
  }

  // return the element name

  String HtmlElement::getname (void) const {
    rdlock ();
    String result = d_name;
    unlock ();
    return result;
  }

  // add a tag to this element

  void HtmlElement::addtag (HtmlTag* tag) {
    if (tag == nilp) return;
    wrlock ();
    d_tags.append (tag);
    unlock ();
  }

  // add a tag to this element by name and value

  void HtmlElement::addtag (const String& name, const Literal& value) {
    HtmlTag* tag = new HtmlTag (name, value);
    addtag (tag);
  }

  // return a opening string of the element
  
  String HtmlElement::opening (void) const {
    rdlock ();
    String result = '<';
    result = result + d_name;
    // add the tags
    long tlen = d_tags.length ();
    for (long i = 0; i < tlen; i++) {
      HtmlTag* tag = dynamic_cast <HtmlTag*> (d_tags.get (i));
      if (tag == nilp) continue;
      result = result + ' ' + tag->tostring ();
    }
    result = result + '>';
    unlock ();
    return result;
  }

  // return a closing string of the element
  
  String HtmlElement::closing (void) const {
    rdlock ();
    String result = "</";
    result = result + d_name + '>';
    unlock ();
    return result;
  }

  // emit this element into a buffer

  void HtmlElement::emit (Buffer& buf) const {
    rdlock  ();
    buf.add (opening ());
    buf.add (closing ());
    buf.add (eolc);
    unlock ();
  }

  // create a new element object in a generic way

  Object* HtmlElement::mknew (Vector* argv) {
    long argc = (argv == nilp) ? 0 : argv->length ();

    // check for one argument
    if (argc == 1) {
      String name = argv->getstring (0);
      return new HtmlElement (name);
    }

    // wrong arguments
    throw Exception ("argument-error", 
		     "invalid arguments with html element object");
  }

  // apply this html element object with a set of arguments and a quark

  Object* HtmlElement::apply (Runnable* robj, Nameset* nset, const long quark,
			      Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_GETNAME) return new String (getname ());
      if (quark == QUARK_OPENING) return new String (opening ());
      if (quark == QUARK_CLOSING) return new String (closing ());
      if (quark == QUARK_RESET) {
	reset ();
	return nilp;
      }
    }

    // dispatch 1 argument
    if (argc == 1) {
      if (quark == QUARK_EMIT) {
	Buffer* buf = dynamic_cast <Buffer*> (argv->get (0));
	if (buf == nilp) {
	  throw Exception ("type-error", "buffer expected with emit");
	}
	emit (*buf);
	return nilp;
      }
      if (quark == QUARK_ADDTAG) {
	HtmlTag* tag = dynamic_cast <HtmlTag*> (argv->get (0));
	if (tag == nilp) {
	  throw Exception ("type-error", "tag expected with add-tag");
	}
	addtag (tag);
	return nilp;
      }
    }

    // dispatch 2 arguments
    if (argc == 2) {
      if (quark == QUARK_ADDTAG) {
	String   name = argv->getstring (0);
	Literal* tval = dynamic_cast <Literal*> (argv->get (1));
	if (tval == nilp) {
	  throw Exception ("type-error", "literal expected with add-tag");
	}
	addtag (name, *tval);
	return nilp;
      }
    }
    // call the object method
    return Object::apply (robj, nset, quark, argv);
  }
}
