// ---------------------------------------------------------------------------
// - HtmlRef.cpp                                                             -
// - afnix:www library - html reference element class implementation         -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2005 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "HtmlRef.hpp"
#include "Runnable.hpp"
#include "Exception.hpp"

namespace afnix {

  // the element supported quarks
  static const long QUARK_GETTEXT = String::intern ("get-text");
  static const long QUARK_SETTEXT = String::intern ("set-text");

  // create a default div element

  HtmlRef::HtmlRef (void) : HtmlElement ("a") {
    d_text = "";
  }

  // create an element with a path

  HtmlRef::HtmlRef (const String& path) : HtmlElement ("a") {
    if (path.length () != 0) addtag ("href", path);
  }

  // create an element with a path and text

  HtmlRef::HtmlRef (const String& path, const String& text) : 
    HtmlElement ("a") {
    if (path.length () != 0) addtag ("href", path);
    d_text= text;
  }

  // return the class name

  String HtmlRef::repr (void) const {
    return "HtmlRef";
  }

  // set the element text
  
  void HtmlRef::settext (const String& text) {
    wrlock ();
    d_text = text;
    unlock ();
  }

  // return the element text

  String HtmlRef::gettext (void) const {
    rdlock ();
    String text = d_text;
    unlock ();
    return text;
  }

  // emit this element into a buffer

  void HtmlRef::emit (Buffer& buf) const {
    rdlock  ();
    try {
      if (d_text.length () > 0) {
	buf.add (opening ());
	buf.add (d_text);
	buf.add (closing ());
	buf.add (eolc);
      }
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // create a new div element object in a generic way

  Object* HtmlRef::mknew (Vector* argv) {
    long argc = (argv == nilp) ? 0 : argv->length ();

    // check for no argument
    if (argc == 0) return new HtmlRef;
    // check for one argument
    if (argc == 1) {
      String path = argv->getstring (0);
      return new HtmlRef (path);
    }
    // check for two arguments
    if (argc == 2) {
      String path = argv->getstring (0);
      String text = argv->getstring (1);
      return new HtmlRef (path, text);
    }
    // wrong arguments
    throw Exception ("argument-error", 
		     "invalid arguments with html reference element object");
  }

  // apply this html div element object with a set of arguments and a quark
  
  Object* HtmlRef::apply (Runnable* robj, Nameset* nset, const long quark,
			  Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_GETTEXT) return new String (gettext ());
    }

    // dispatch 1 argument
    if (argc == 1) {
      if (quark == QUARK_SETTEXT) {
	String text = argv->getstring (0);
	settext (text);
	return nilp;
      }
    }
    // call the element method
    return HtmlElement::apply (robj, nset, quark, argv);
  }
}
