// ---------------------------------------------------------------------------
// - HtmlTable.cpp                                                           -
// - afnix:www library - html table class implementation                     -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2005 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Vector.hpp"
#include "Integer.hpp"
#include "Runnable.hpp"
#include "HtmlTable.hpp"
#include "Exception.hpp"

namespace afnix {

  // -------------------------------------------------------------------------
  // - table data object                                                     -
  // -------------------------------------------------------------------------

  // the table data supported quarks
  static const long QUARK_SETDATA  = String::intern ("set-data");
  static const long QUARK_GETDATA  = String::intern ("get-data");

  // create an empty table data

  HtmlTd::HtmlTd (void) : HtmlElement ("td") {
    d_data = "&nbsp;";
  }

  // create a table data element by value

  HtmlTd::HtmlTd (const Literal& lval) : HtmlElement ("td") {
    d_data = lval.tostring ();
    if (d_data.length () == 0) d_data = "&nbsp;";
  }

  // return the class name

  String HtmlTd::repr (void) const {
    return "HtmlTd";
  }

  // set the table data element by literal

  void HtmlTd::setdata (const Literal& lval) {
    wrlock ();
    d_data = lval.tostring ();
    if (d_data.length () == 0) d_data = "&nbsp;";
    unlock ();
  }

  // return the table data element

  String HtmlTd::getdata (void) const{
    rdlock ();
    String result = d_data;
    unlock ();
    return result;
  }

  // emit this table data into a buffer

  void HtmlTd::emit (Buffer& buf) const {
    rdlock  ();
    buf.add (opening ());
    buf.add (d_data);
    buf.add (closing ()); 
    buf.add (eolc);
    unlock ();
  }

  // create a new table data in a generic way

  Object* HtmlTd::mknew (Vector* argv) {
    long argc = (argv == nilp) ? 0 : argv->length ();
    // check for 0 argument
    if (argc == 0) return new HtmlTd;
    // check for 1 argument
    if (argc == 1) {
      Literal* lval = dynamic_cast <Literal*> (argv->get (0));
      if (lval == nilp)
	throw Exception ("argument-error",
			 "invalid argument with table data object");
      return new HtmlTd (*lval);
    }
    // wrong arguments
    throw Exception ("argument-error", 
		     "too many arguments with html table data object");
  }

  // apply this table data object with a set of arguments and a quark

  Object* HtmlTd::apply (Runnable* robj, Nameset* nset, const long quark, 
			 Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // dispqtch 0 argument
    if (argc == 0) {
      if (quark == QUARK_GETDATA) return new String (getdata ());
    }

    // dispatch 1 argument
    if (argc == 1) {
      if (quark == QUARK_SETDATA) {
	Literal* lval = dynamic_cast <Literal*> (argv->get (0));
	if (lval == nilp)
	  throw Exception ("type-error", "literal expected with set-data");
	setdata (*lval);
	return nilp;
      }
    }

    // call the html element method
    return HtmlElement::apply (robj, nset, quark, argv);
  }

  // -------------------------------------------------------------------------
  // - table row object                                                     -
  // -------------------------------------------------------------------------

  // the table data supported quarks
  static const long QUARK_LENGTH   = String::intern ("length");
  static const long QUARK_ADDDATA  = String::intern ("add-data");

  // create an empty table row

  HtmlTr::HtmlTr (void) : HtmlElement ("tr") {
    d_tbld.reset ();
  }

  // return the class name

  String HtmlTr::repr (void) const {
    return "HtmlTr";
  }

  // return the number of data elements

  long HtmlTr::length (void) const {
    rdlock ();
    long result = d_tbld.length ();
    unlock ();
    return result;
  }

  // add a table data element to the row

  void HtmlTr::adddata (HtmlTd* td) {
    d_tbld.append (td);
  }

  // add a data literal to the row

  void HtmlTr::adddata (const Literal& lval) {
    d_tbld.append (new HtmlTd (lval));
  }

  // return the table data element

  HtmlTd* HtmlTr::getdata (const long index) const{
    rdlock ();
    try {
      HtmlTd* result = dynamic_cast <HtmlTd*> (d_tbld.get (index));
      unlock ();
      return result;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // emit this table row into a buffer

  void HtmlTr::emit (Buffer& buf) const {
    rdlock  ();
    buf.add (opening ()); buf.add (eolc);
    long tlen = d_tbld.length ();
    for (long i = 0; i < tlen; i++) {
      HtmlTd* td = dynamic_cast <HtmlTd*> (d_tbld.get (i));
      if (td != nilp) td->emit (buf);
    }
    buf.add (closing ()); buf.add (eolc);
    unlock ();
  }

  // create a new table row in a generic way

  Object* HtmlTr::mknew (Vector* argv) {
    long argc = (argv == nilp) ? 0 : argv->length ();
    // check for 0 argument
    if (argc == 0) return new HtmlTr;
    // wrong arguments
    throw Exception ("argument-error", 
		     "too many arguments with html table row object");
  }

  // apply this table data object with a set of arguments and a quark

  Object* HtmlTr::apply (Runnable* robj, Nameset* nset, const long quark, 
			 Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_LENGTH) return new Integer (length ());
    }

    // dispatch 1 argument
    if (argc == 1) {
      if (quark == QUARK_ADDDATA) {
	Object* obj = argv->get (0);
	HtmlTd* td = dynamic_cast <HtmlTd*> (obj);
	if (td != nilp) {
	  adddata (td);
	  return nilp;
	}
	Literal* lval = dynamic_cast <Literal*> (obj);
	if (lval != nilp) {
	  adddata (*lval);
	  return nilp;
	}
	throw Exception ("type-error", "unexpected object with add-data",
			 Object::repr (obj));
      }
      if (quark == QUARK_GETDATA) {
	rdlock ();
	try {
	  long index = argv->getint (0);
	  HtmlTd* td = getdata (index);
	  robj->post (td);
	  return td;
	} catch (...) {
	  unlock ();
	  throw;
	}
      }
    }

    // call the html element method
    return HtmlElement::apply (robj, nset, quark, argv);
  }

  // -------------------------------------------------------------------------
  // - html table object                                                     -
  // -------------------------------------------------------------------------

  // the html table supported quarks
  static const long QUARK_ADDHEAD    = String::intern ("add-head");
  static const long QUARK_ADDBODY    = String::intern ("add-body");
  static const long QUARK_ADDFOOT    = String::intern ("add-foot");
  static const long QUARK_ADDTABLE   = String::intern ("add-table");
  static const long QUARK_HEADLENGTH = String::intern ("head-length");
  static const long QUARK_BODYLENGTH = String::intern ("body-length");
  static const long QUARK_FOOTLENGTH = String::intern ("foot-length");
  static const long QUARK_GETCAPTION = String::intern ("get-caption");
  static const long QUARK_SETCAPTION = String::intern ("set-caption");

  // create a new html table

  HtmlTable::HtmlTable (void) : HtmlElement ("table") {
    d_thrs.reset ();
    d_tbrs.reset ();
    d_tfrs.reset ();
    d_caption = "";
  }

  // create a new html table with a print table

  HtmlTable::HtmlTable (const PrintTable& ptbl) : HtmlElement ("table") {
    d_thrs.reset ();
    d_tbrs.reset ();
    d_tfrs.reset ();
    d_caption = "";
    addtbl (ptbl);
  }

  // return the class name

  String HtmlTable::repr (void) const {
    return "HtmlTable";
  }

  // set the table caption

  void HtmlTable::setcaption (const String& caption) {
    wrlock ();
    d_caption = caption;
    unlock ();
  }

  // return the table caption

  String HtmlTable::getcaption (void) const {
    rdlock ();
    String result = d_caption;
    unlock ();
    return result;
  }

  // return the number of head rows

  long HtmlTable::gethlen (void) const {
    rdlock ();
    long result = d_thrs.length ();
    unlock ();
    return result;
  }

  // return the number of body rows

  long HtmlTable::getblen (void) const {
    rdlock ();
    long result = d_tbrs.length ();
    unlock ();
    return result;
  }

  // return the number of foot rows

  long HtmlTable::getflen (void) const {
    rdlock ();
    long result = d_tfrs.length ();
    unlock ();
    return result;
  }

  // add a row to the head table

  void HtmlTable::addhead (HtmlTr* tr) {
    if (tr == nilp) return;
    wrlock ();
    try {
      d_thrs.append (tr);
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // add a row to the body table

  void HtmlTable::addbody (HtmlTr* tr) {
    if (tr == nilp) return;
    wrlock ();
    try {
      d_tbrs.append (tr);
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // add a row to the foot table

  void HtmlTable::addfoot (HtmlTr* tr) {
    if (tr == nilp) return;
    wrlock ();
    try {
      d_tfrs.append (tr);
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // add a print table to the table

  void HtmlTable::addtbl (const PrintTable& ptbl) {
    // get the number of rows and columns
    long rows = ptbl.getrows ();
    long cols = ptbl.getcols ();
    if ((rows == 0) || (cols == 0)) return;
    wrlock ();
    try {
      for (long i = 0; i < rows; i++) {
	HtmlTr* tr = new HtmlTr;
	for (long j = 0; j < cols; j++) {
	  String data = ptbl.get (i, j);
	  tr->adddata (data);
	}
	addbody (tr);
      }
    } catch (...) {
      unlock ();
      throw;
    }
    unlock ();
  }

  // emit this table row into a buffer

  void HtmlTable::emit (Buffer& buf) const {
    rdlock  ();
    // open the table
    buf.add (opening ()); buf.add (eolc);
    // emit the caption
    if (d_caption.length () > 0) {
      buf.add (String ("<caption>"));
      buf.add (d_caption);
      buf.add (String ("</caption>\n"));
    }
    // emit the head
    long hlen = d_thrs.length ();
    if (hlen > 0) {
      buf.add (String ("<thead>\n"));
      for (long i = 0; i < hlen; i++) {
	HtmlTr* tr = dynamic_cast <HtmlTr*> (d_thrs.get (i));
	if (tr != nilp) tr->emit (buf);
      }
      buf.add (String ("</thead>\n"));
    }
    // emit the body
    long blen = d_tbrs.length ();
    if (blen > 0) {
      buf.add (String ("<tbody>\n"));
      for (long i = 0; i < blen; i++) {
	HtmlTr* tr = dynamic_cast <HtmlTr*> (d_tbrs.get (i));
	if (tr != nilp) tr->emit (buf);
      }
      buf.add (String ("</tbody>\n"));
    }
    // emit the foot
    long flen = d_tfrs.length ();
    if (flen > 0) {
      buf.add (String ("<tfoot>\n"));
      for (long i = 0; i < flen; i++) {
	HtmlTr* tr = dynamic_cast <HtmlTr*> (d_tfrs.get (i));
	if (tr != nilp) tr->emit (buf);
      }
      buf.add (String ("</tfoot>\n"));
    }
    // close the table
    buf.add (closing ()); buf.add (eolc);
    unlock ();
  }

  // create a new html table in a generic way

  Object* HtmlTable::mknew (Vector* argv) {
    long argc = (argv == nilp) ? 0 : argv->length ();
    // check for 0 argument
    if (argc == 0) return new HtmlTable;
    // check for 1 argument
    if (argc == 1) {
      PrintTable* ptbl = dynamic_cast <PrintTable*> (argv->get (0));
      if (ptbl == nilp) 
	throw Exception ("type-error", "invalid argmument with html table");
      return new HtmlTable (*ptbl);
    }
    // wrong arguments
    throw Exception ("argument-error", 
		     "too many arguments with html table object");
  }

  // apply this html table object with a set of arguments and a quark

  Object* HtmlTable::apply (Runnable* robj, Nameset* nset, const long quark,
			    Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();
 
    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_HEADLENGTH) return new Integer (gethlen ());
      if (quark == QUARK_BODYLENGTH) return new Integer (getblen ());
      if (quark == QUARK_FOOTLENGTH) return new Integer (getflen ());
      if (quark == QUARK_GETCAPTION) return new String  (getcaption ());
    }
	
    // dispatch 1 argument
    if (argc == 1) {
      if (quark == QUARK_SETCAPTION) {
	String caption = argv->getstring (0);
	setcaption (caption);
	return nilp;
      }
      if (quark == QUARK_ADDHEAD) {
	HtmlTr* tr = dynamic_cast <HtmlTr*> (argv->get (0));
	if (tr == nilp)
	  throw Exception ("type-error", "row element expected with add-head");
	addhead (tr);
	return nilp;
      }
      if (quark == QUARK_ADDBODY) {
	HtmlTr* tr = dynamic_cast <HtmlTr*> (argv->get (0));
	if (tr == nilp)
	  throw Exception ("type-error", "row element expected with add-body");
	addbody (tr);
	return nilp;
      }
      if (quark == QUARK_ADDFOOT) {
	HtmlTr* tr = dynamic_cast <HtmlTr*> (argv->get (0));
	if (tr == nilp)
	  throw Exception ("type-error", "row element expected with add-foot");
	addfoot (tr);
	return nilp;
      }
      if (quark == QUARK_ADDTABLE) {
	PrintTable* tbl = dynamic_cast <PrintTable*> (argv->get (0));
	if (tbl == nilp)
	  throw Exception ("type-error", "print table expected for add-table");
	addtbl (*tbl);
	return nilp;
      }
    }
    // call the html element method
    return HtmlElement::apply (robj, nset, quark, argv);
  }
}
