// ---------------------------------------------------------------------------
// - HtmlTag.cpp                                                             -
// - afnix:www library - html tag class implementation                       -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2005 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Vector.hpp"
#include "HtmlTag.hpp"
#include "Exception.hpp"

namespace afnix {
  // the tag supported quarks
  static const long QUARK_RESET    = String::intern ("reset");
  static const long QUARK_GETNAME  = String::intern ("get-name");
  static const long QUARK_SETNAME  = String::intern ("set-name");
  static const long QUARK_GETVALUE = String::intern ("get-value");
  static const long QUARK_SETVALUE = String::intern ("set-value");
  static const long QUARK_TOSTRING = String::intern ("to-string");

  // create a default tag

  HtmlTag::HtmlTag (void) {
    d_name = "";
    d_tval = "";
  }

  // create a tag by name

  HtmlTag::HtmlTag (const String& name) {
    d_name = name;
    d_tval = name;
  }

  // create a tag by name and value

  HtmlTag::HtmlTag (const String& name, const Literal& tval) {
    d_name = name;
    d_tval = tval.tostring ();
  }

  // copy construct this html tag

  HtmlTag::HtmlTag (const HtmlTag& that) {
    that.rdlock ();
    d_name = that.d_name;
    d_tval = that.d_tval;
    that.unlock ();
  }

  // assign an html tag to this one

  HtmlTag& HtmlTag::operator = (const HtmlTag& that) {
    // check for equality
    if (this == &that) return *this;
    // lock and assign
    wrlock ();
    that.rdlock ();
    d_name = that.d_name;
    d_tval = that.d_tval;
    that.unlock ();
    unlock ();
    return *this;
  }

  // return the class name

  String HtmlTag::repr (void) const {
    return "HtmlTag";
  }

  // reset this tag

  void HtmlTag::reset (void) {
    wrlock ();
    d_name   = "";
    d_tval   = "";
    unlock ();
  }

  // set the tag name

  void HtmlTag::setname (const String& name) {
    wrlock ();
    d_name = name;
    unlock ();
  }

  // return the tag name

  String HtmlTag::getname (void) const {
    rdlock ();
    String result = d_name;
    unlock ();
    return result;
  }

  // set the tag value

  void HtmlTag::setvalue (const Literal& tval) {
    wrlock ();
    d_tval = tval.tostring ();
    unlock ();
  }

  // return the tag value

  String HtmlTag::getvalue (void) const {
    rdlock ();
    String result = d_tval;
    unlock ();
    return result;
  }

  // return a string representation
  
  String HtmlTag::tostring (void) const {
    rdlock ();
    String result = d_name;
    result = result + "=" + '"' + d_tval + '"';
    unlock ();
    return result;
  }

  // create a new tag object in a generic way

  Object* HtmlTag::mknew (Vector* argv) {
    long argc = (argv == nilp) ? 0 : argv->length ();

    // check for 0 argument
    if (argc == 0) return new HtmlTag;

    // check for one argument
    if (argc == 1) {
      String name = argv->getstring (0);
      return new HtmlTag (name);
    }

    // check for two arguments
    if (argc == 2) {
      String   name = argv->getstring (0);
      Literal* tval = dynamic_cast <Literal*> (argv->get (1));
      if (tval == nilp) {
	throw Exception ("argument-error", 
			 "invalid arguments with html tag object");
      }
      return new HtmlTag (name, *tval);
    }
    // wrong arguments
    throw Exception ("argument-error", 
		     "invalid arguments with html tag object");
  }

  // apply this html tag object with a set of arguments and a quark

  Object* HtmlTag::apply (Runnable* robj, Nameset* nset, const long quark,
			  Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_GETNAME)  return new String (getname  ());
      if (quark == QUARK_GETVALUE) return new String (getvalue ());
      if (quark == QUARK_TOSTRING) return new String (tostring ());
      if (quark == QUARK_RESET) {
	reset ();
	return nilp;
      }
    }

    // dispatch 1 argument
    if (argc == 1) {
      if (quark == QUARK_SETNAME) {
	String name = argv->getstring (0);
	setname (name);
	return nilp;
      }
      if (quark == QUARK_SETVALUE) {
	String   name = argv->getstring (0);
	Literal* tval = dynamic_cast <Literal*> (argv->get (0));
	if (tval == nilp) {
	  throw Exception ("argument-error", 
			   "invalid arguments with html tag object");
	}
	setvalue (*tval);
	return nilp;
      }
    }
    // call the object method
    return Object::apply (robj, nset, quark, argv);
  }
}
